

import copy


import numpy as np
import scipy.interpolate


# ################################################################
# ################################################################
#                     _DCRYST_MAT
# ################################################################
# ################################################################


_DCRYST_MAT = {

    # ------
    # Quartz

    'Quartz': {
        'material_symbol': 'Qz',
        'atoms': ['Si', 'O'],
        'atoms_Z': [14., 8.],
        'atoms_nb': [3., 6.],
        'volume': None,
        'mesh': {
            'type': 'hexagonal',
            'positions': {
                'Si': {
                    'u': np.r_[0.465],
                    'x': None,
                    'y': None,
                    'z': None,
                    'N': None,
                },
                'O': {
                    'u': np.r_[0.415, 0.272, 0.120],
                    'x': None,
                    'y': None,
                    'z': None,
                    'N': None,
                },
            },
            'sources': 'R.W.G. Wyckoff, Crystal Structures (1963)',
        },
        'inter_atomic': {
            'distances': {
                'a0': 4.91304, # e-10,
                'c0': 5.40463, # e-10,
            },
            'unit': 'A',
            'comments': 'within the unit cell',
            'Tref': {
                'data': 25. + 273.15,
                'unit': 'K',
            },
            'sources': 'R.W.G. Wyckoff, Crystal Structures',
        },
        'phases': { # Will be populated
            'Si': None,
            'O': None,
        },
        'thermal_expansion': {
            'coefs': {
                'alpha_a': 1.337e-5,
                'alpha_c': 7.97e-6,
            },
            'unit': '1/K',
            'comments': 'in parallel directions to a0 and c0',
            'sources': 'R.W.G. Wyckoff, Crystal Structures',
        },
        'sin_theta_lambda': {
            'Si': np.r_[
                0., 0.1, 0.2, 0.25, 0.3, 0.35, 0.4, 0.5, 0.6, 0.7,
                0.8, 0.9, 1., 1.1, 1.2, 1.3, 1.4, 1.5,
            ], # *1e10,
            'O': np.r_[
                0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1., 1.1,
            ], # *1e10,
            'sources':
                'Int. Tab. X-Ray Crystallography, Vol.I,II,III,IV (1985)',
        },
        'atomic_scattering': {
            'factors': {
                'Si': np.r_[
                    12., 11., 9.5, 8.8, 8.3, 7.7, 7.27, 6.25, 5.3,
                    4.45, 3.75, 3.15, 2.7, 2.35, 2.07, 1.87, 1.71, 1.6,
                ],
                'O': np.r_[
                    9., 7.836, 5.756, 4.068, 2.968, 2.313, 1.934, 1.710, 1.566,
                    1.462, 1.373, 1.294,
                ],
            },
            'sources':
                'Int. Tab. X-Ray Crystallography, Vol.I,II,III,IV (1985)',
        },
    },

    # ----------
    # Germanium

    'Germanium': {
        'material_symbol': 'Ge',
        'atoms': ['Ge'],
        'atoms_Z': [32.],
        'atoms_nb': [8.],
        'volume': None,
        'mesh': {
            'type': 'diamond',
            'positions': {
                'Ge': {
                    'u': np.r_[0.25],
                    'x': None,
                    'y': None,
                    'z': None,
                    'N': None,
                },
            },
            'sources':  'R.W.G. Wyckoff, Crystal Structures (1963), Eqn 8a',
        },
        'inter_atomic': {
            'distances': {
                'a0': 5.65735, # e-10
            },
            'unit': 'A',
            'comments': None,
            'Tref': {
                'data': 20. + 273.15,
                'unit': 'K',
            },
            'sources': 'R.W.G. Wyckoff, Crystal Structures (1963), Table II.5',
        },
        'phases': { # Will be populated
            'Ge': None,
        },
        'thermal_expansion': {
            'coefs': {
                'alpha_a': 6.12e-6
            },
            'unit': '1/K',
            'comments': 'only one unique direction',
            'sources': 'H.P. Singh, Acta Cryst. (1968). A24, 469',
        },
        'sin_theta_lambda': {
            'Ge': np.r_[
                0., 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35,
                0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1., 1.1, 1.2,
                1.3, 1.4, 1.5
            ], # [1/AA],
            'sources':
                'Int. Tab. X-Ray Crystallography, Vol.I,II,III,IV (1985), Table 3.3.1A',
        },
        'atomic_scattering': {
            'factors': {
                'Ge': np.r_[
                    32.0, 31.28, 29.52, 27.48, 25.53, 23.76,
                    22.11, 20.54, 19.02, 16.19, 13.72, 11.68,
                    10.08, 8.87, 7.96, 7.29, 6.77, 6.37,
                    6.02, 5.72
                ]
            },
            'sources':
                'Int. Tab. X-Ray Crystallography, Vol.I,II,III,IV (1985), Table 3.3.1A',
        },
    },
}


# ##############################################################
# ##############################################################
#           Complement materials dict
# ##############################################################


def _complement_dict_mat(dcryst_mat=None):
    """ Add what is missing to material dict """

    for k0, v0 in dcryst_mat.items():

        # ---------
        # positions
        # ---------

        # Quartz
        if k0 == 'Quartz':
            _positions_quartz(
                dcryst_mat=dcryst_mat,
                k0=k0,
                v0=v0,
            )

        # Germanium
        elif k0 == 'Germanium':
            _positions_germanium(
                dcryst_mat=dcryst_mat,
                k0=k0,
                v0=v0,
            )

        # other
        else:
            msg = "Unknown material"
            raise NotImplementedError(msg)

        # ------
        # volume
        # ------

        # hexagonal
        if v0['mesh']['type'] == 'hexagonal':
            dcryst_mat[k0]['volume'] = hexa_volume(
                v0['inter_atomic']['distances']['a0'],
                v0['inter_atomic']['distances']['c0'],
            )

        # diamond
        elif v0['mesh']['type'] == 'diamond':
            dcryst_mat[k0]['volume'] = diam_volume(
                v0['inter_atomic']['distances']['a0']
            )

        # other
        else:
            msg = f"Mesh volume not implemented for '{k0}'"
            raise NotImplementedError(msg)

        # -----------------------------------------------------
        # atomic absorption coefficients and scattering factor
        # -----------------------------------------------------

        if k0 == 'Quartz':
            _atomic_coefs_factor_Quartz(
                dcryst_mat=dcryst_mat,
                k0=k0,
                v0=v0,
            )

        elif k0 == 'Germanium':
            _atomic_coefs_factor_Germanium(
                dcryst_mat=dcryst_mat,
                k0=k0,
                v0=v0,
            )

        # other
        else:
            msg = "Unknown material"
            raise NotImplementedError(msg)


# ###############################################################
# ###############################################################
#                   Positions
# ##############################################################
# ###############################################################


def _positions_quartz(
    dcryst_mat=None,
    k0=None,
    v0=None,
):

    # Si and O positions for alpha-Quartz crystal
    # From R.W.G. Wyckoff, Crystal Structures (1963)
    # xsi = np.r_[-u, u, 0.]
    # ysi = np.r_[-u, 0., u]
    # zsi = np.r_[1./3., 0., 2./3.]
    # xo = np.r_[x, y - x, -y, x - y, y, -x]
    # yo = np.r_[y, -x, x - y, -y, x, y - x]
    # zo = np.r_[z, z + 1./3., z + 2./3., -z, 2./3. - z, 1./3. - z]

    # Quartz
    uSi = v0['mesh']['positions']['Si']['u'][0]
    dcryst_mat[k0]['mesh']['positions']['Si']['x'] = np.r_[
        -uSi,
        uSi,
        0.
    ]
    dcryst_mat[k0]['mesh']['positions']['Si']['y'] = np.r_[
        -uSi,
        0.,
        uSi
    ]
    dcryst_mat[k0]['mesh']['positions']['Si']['z'] = np.r_[
        1./3.,
        0.,
        2./3.
    ]
    dcryst_mat[k0]['mesh']['positions']['Si']['N'] = np.size(
        dcryst_mat[k0]['mesh']['positions']['Si']['x']
    )

    # Oxygen
    uOx = v0['mesh']['positions']['O']['u'][0]
    uOy = v0['mesh']['positions']['O']['u'][1]
    uOz = v0['mesh']['positions']['O']['u'][2]
    dcryst_mat[k0]['mesh']['positions']['O']['x'] = np.r_[
        uOx,
        uOy - uOx,
        -uOy,
        uOx - uOy,
        uOy,
        -uOx
    ]
    dcryst_mat[k0]['mesh']['positions']['O']['y'] = np.r_[
        uOy,
        -uOx,
        uOx - uOy,
        -uOy,
        uOx,
        uOy - uOx
    ]
    dcryst_mat[k0]['mesh']['positions']['O']['z'] = np.r_[
        uOz,
        uOz + 1./3.,
        uOz + 2./3.,
        -uOz,
        2./3. - uOz,
        1./3. - uOz
    ]
    dcryst_mat[k0]['mesh']['positions']['O']['N'] = np.size(
        dcryst_mat[k0]['mesh']['positions']['O']['x']
    )


def _positions_germanium(
    dcryst_mat=None,
    k0=None,
    v0=None,
):

    # Ge positions for Ge crystal
    # From R.W.G. Wyckoff, Crystal Structures (1963)

    # Germanium
    uGe = v0['mesh']['positions']['Ge']['u'][0]
    dcryst_mat[k0]['mesh']['positions']['Ge']['x'] = np.r_[
        0.,
        0.,
        2*uGe,
        2*uGe,
        uGe,
        uGe,
        3*uGe,
        3*uGe
    ]
    dcryst_mat[k0]['mesh']['positions']['Ge']['y'] = np.r_[
        0.,
        2*uGe,
        0.,
        2*uGe,
        uGe,
        3*uGe,
        uGe,
        3*uGe
    ]
    dcryst_mat[k0]['mesh']['positions']['Ge']['z'] = np.r_[
        0.,
        2*uGe,
        2*uGe,
        0.,
        uGe,
        3*uGe,
        3*uGe,
        uGe

    ]
    dcryst_mat[k0]['mesh']['positions']['Ge']['N'] = np.size(
        dcryst_mat[k0]['mesh']['positions']['Ge']['x']
    )



# ##################################################################
# ##################################################################
#                   Elementary box volume
# ##################################################################
# ##################################################################


def hexa_volume(aa, cc):
    return (aa**2) * cc * (np.sqrt(3.)/2.)

def diam_volume(aa):
    return aa**3


# ###############################################################
# ###############################################################
#       atomic absorption coefficients and scattering factor
# ##############################################################
# ###############################################################


def _atomic_coefs_factor_Quartz(
    dcryst_mat=None,
    k0=None,
    v0=None,
):

    Zsi, Zo = v0['atoms_Z']

    # -----------------------------------
    # linear atomic absorption coefficients 'mu'
    # From W. Zachariasen, Theory of X-ray Diffraction in Crystals
    # (Wiley, New York, 1945)

    # Data file for anomalous corrections to atomic scattering factor & attenuation
    # From Lawrence Berkeley National Lab, Center for X-Ray Optics
    # https://henke.lbl.gov/optical_constants/asf.html

    def mu_si(lamb, Zsi=Zsi):
        if lamb > 6.74:     # e-10 ?
            return 5.33e-4*(lamb**2.74)*(Zsi**3.03)
        else:
            return 1.38e-2*(lamb**2.79)*(Zsi**2.73)

    def mu_o(lamb, Z0=Zo):
        return 5.4e-3*(lamb**2.92)*(Zo**3.07)

    def mu(lamb, mu_si=mu_si, mu_o=mu_o):
        return 2.65e-8*(7.*mu_si(lamb) + 8.*mu_o(lamb))/15.

    # store in dict
    dcryst_mat[k0]['mu'] = mu

    # ----------------------------
    # Atomic scattering factor 'f'

    # Same values for 110- and Quartz_102
    sol_si = v0['sin_theta_lambda']['Si']
    sol_o = v0['sin_theta_lambda']['O']
    asf_si = v0['atomic_scattering']['factors']['Si']
    asf_o = v0['atomic_scattering']['factors']['O']
    interp_si = scipy.interpolate.interp1d(sol_si, asf_si)
    interp_o = scipy.interpolate.interp1d(sol_o, asf_o)

    def dfsi_re(lamb):
        return 0.1335*lamb - 6e-3

    def fsi_re(lamb, sol, dfsi_re=dfsi_re, interp_si=interp_si):
        return interp_si(sol) + dfsi_re(lamb)

    def fsi_im(lamb, Zsi=Zsi, mu_si=mu_si):
        return 5.936e-4*Zsi*(mu_si(lamb)/lamb)

    def dfo_re(lamb):
        return 0.1335*lamb - 0.206

    def fo_re(lamb, sol, dfo_re=dfo_re, interp_o=interp_o):
        return interp_o(sol) + dfo_re(lamb)

    def fo_im(lamb, Zo=Zo, mu_o=mu_o):
        return 5.936e-4*Zo*(mu_o(lamb)/lamb)

    # store in dict
    dcryst_mat[k0]['dfsi_re'] = dfsi_re
    dcryst_mat[k0]['dfo_re'] = dfo_re
    dcryst_mat[k0]['fsi_re'] = fsi_re
    dcryst_mat[k0]['fsi_im'] = fsi_im
    dcryst_mat[k0]['fo_re'] = fo_re
    dcryst_mat[k0]['fo_im'] = fo_im

def _atomic_coefs_factor_Germanium(
    dcryst_mat=None,
    k0=None,
    v0=None,
):

    # Useful constant
    import scipy.constants as cnt
    hc = cnt.h*cnt.c/cnt.e*1e10 # [eV*AA]

    # Charge of Germanium
    Zge = v0['atoms_Z']

    # Density of Germanium
    rho_Ge = 5.307 # [g/cm^3]

    # Data file for anomalous corrections to atomic scattering factor & attenuation
    # From Lawrence Berkeley National Lab, Center for X-Ray Optics
    # https://henke.lbl.gov/optical_constants/asf.html

    import os
    table_path = os.path.join(
        os.path.dirname(__file__),
        'Ge_AtomicScatteringFactors_LBL.txt'
        )

    tables = np.loadtxt(table_path)

    # Table values
    E_LBL  = tables[:,0] # [eV]
    f1_LBL = tables[:,1] # [e/atom]
    f2_LBL = tables[:,2] # [e/atom]

    # ----------------------------
    # Atomic scattering factor 'f'

    # Interpolates mean scattering factors
    sol_ge = v0['sin_theta_lambda']['Ge'] # [1/AA]
    asf_ge = v0['atomic_scattering']['factors']['Ge'] # [e/atom]
    interp_ge_f0 = scipy.interpolate.interp1d(sol_ge, asf_ge)

    # Interpolates NIST anomalous scattering factor corrections
    interp_ge_f1 = scipy.interpolate.interp1d(E_LBL, f1_LBL)
    interp_ge_f2 = scipy.interpolate.interp1d(E_LBL, f2_LBL)

    def dfge_re(lamb, Zge=Zge):
        return interp_ge_f1(hc/lamb) - Zge

    def fge_re(lamb, sol, dfge_re=dfge_re, interp_ge_f0=interp_ge_f0):
        return interp_ge_f0(sol) + dfge_re(lamb)

    def fge_im(lamb):
        return interp_ge_f2(hc/lamb)


    # store in dict
    dcryst_mat[k0]['dfge_re'] = dfge_re
    dcryst_mat[k0]['fge_re'] = fge_re
    dcryst_mat[k0]['fge_im'] = fge_im

    # -----------------------------------
    # linear atomic absorption coefficients 'mu'

    # Material-dependent conversion from f2 to mu
    # From NIST X-ray Form Factor, Attenuation, and Scattering Tables
    # https://physics.nist.gov/PhysRefData/FFast/html/form.html
    def mu_ge(lamb):
        return 5.7969e5*fge_im(lamb)*(lamb/hc) * rho_Ge *1e-8 # [1/AA]

    # store in dict
    dcryst_mat[k0]['mu'] = mu_ge



# ###############################################################
# ###############################################################
#                     _DCRYST
# ##############################################################
# ###############################################################


_DCRYST = {
    'Quartz_110': {
        'material': 'Quartz',
        'name': 'Quartz_110',
        'symbol': 'Qz110',
        'miller': np.r_[1., 1., 0.],
        'target': {
            'ion': 'Ar16+',
            'lamb': 3.96, # e-10,
            'units': 'm',
        },
        'd_hkl': None,
    },
    'Quartz_102': {
        'material': 'Quartz',
        'name': 'Quartz_102',
        'symbol': 'Qz102',
        'miller': np.r_[1., 0., 2.],
        'target': {
            'ion': 'Ar17+',
            'lamb': 3.75, # e-10,
            'units': 'm',
        },
        'd_hkl': None,
    },

    # 'Germanium_XXX': {
        # 'material': 'Germanium',
        # 'name': None,
        # 'symbol': None,
        # 'miller': None,
        # 'target': {
            # 'ion': 'Ar16+',
            # 'wavelength': 3.96e-10,
        # }
        # 'd_hkl': None,
    # },
}


# ##############################################################
# ##############################################################
#           Complement cut dict
# ##############################################################


def _complement_dict_cut(dcryst_mat=None, dcryst_cut=None):
    """ Add what is missing to material dict """

    for k0, v0 in dcryst_cut.items():

        # ------------------------
        # complement with material

        # safety check on material
        if v0.get('material') not in dcryst_mat.keys():
            lstr = [f"\t- '{k0}'" for k0 in dcryst_mat.keys()]
            msg = (
                f"Please, for '{k0}', select a material from:\n"
                + "\n".join(lstr)
                + f"\nProvided: {v0.get('material')}\n"
            )
            raise Exception(msg)

        # complement dict
        for k1, v1 in dcryst_mat[v0['material']].items():
            dcryst_cut[k0][k1] = copy.deepcopy(v1)

        # -----------
        # preparation

        hh, kk, ll = v0['miller']

        # ------------------------------
        # inter-reticular distance d_hkl

        # hexagonal
        if v0['mesh']['type'] == 'hexagonal':
            dcryst_cut[k0]['d_hkl'] = hexa_spacing(
                hh,
                kk,
                ll,
                v0['inter_atomic']['distances']['a0'],
                v0['inter_atomic']['distances']['c0'],
            )

        # diamond
        elif v0['mesh']['type'] == 'diamond':
            dcryst_cut[k0]['d_hkl'] = diam_spacing(
                hh,
                kk,
                ll,
                v0['inter_atomic']['distances']['a0'],
            )

        # other
        else:
            msg = f"Mesh d_hkl not implemented for '{k0}'"
            raise NotImplementedError(msg)

        # ------------
        # phases

        # Quartz
        if v0['material'] == 'Quartz':

            Nsi = v0['mesh']['positions']['Si']['N']
            No = v0['mesh']['positions']['O']['N']

            def phasesi(hh, kk, ll, xsi, ysi, zsi):
                return hh*xsi + kk*ysi + ll*zsi

            def phaseo(hh, kk, ll, xo, yo, zo):
                return hh*xo + kk*yo + ll*zo

            # initiate
            dcryst_cut[k0]['phases']['Si'] = np.full((Nsi,), np.nan)
            dcryst_cut[k0]['phases']['O'] = np.full((No,), np.nan)

            # Quartz
            for ii in range(Nsi):
                dcryst_cut[k0]['phases']['Si'][ii] = phasesi(
                    hh,
                    kk,
                    ll,
                    v0['mesh']['positions']['Si']['x'][ii],
                    v0['mesh']['positions']['Si']['y'][ii],
                    v0['mesh']['positions']['Si']['z'][ii],
                )

            # Oxygen
            for ii in range(No):
                dcryst_cut[k0]['phases']['O'][ii] = phaseo(
                    hh,
                    kk,
                    ll,
                    v0['mesh']['positions']['O']['x'][ii],
                    v0['mesh']['positions']['O']['y'][ii],
                    v0['mesh']['positions']['O']['z'][ii],
                )

        # Germanium
        elif v0['material'] == 'Germanium':

            Nge = v0['mesh']['positions']['Ge']['N']

            # dot(s, r_atom); s = lattice vector, r_atom = atom position
            def phasege(hh, kk, ll, xge, yge, zge):
                return hh*xge + kk*yge + ll*zge

            # initiate
            dcryst_cut[k0]['phases']['Ge'] = np.full((Nge,), np.nan)

            # Germanium
            for ii in range(Nge):
                dcryst_cut[k0]['phases']['Ge'][ii] = phasege(
                    hh,
                    kk,
                    ll,
                    v0['mesh']['positions']['Ge']['x'][ii],
                    v0['mesh']['positions']['Ge']['y'][ii],
                    v0['mesh']['positions']['Ge']['z'][ii],
                )

        # Other
        else:
            msg = f"Phases not implemented for '{k0}'"
            raise NotImplementedError(msg)

    return dcryst_cut


# #################################################################
# #################################################################
#               Inter-reticular spacing d_hkl
# #################################################################
# #################################################################


def hexa_spacing(hh, kk, ll, aa, cc):
    return np.sqrt(
        (3.*(aa**2)*(cc**2))
        / (4.*(hh**2 + kk**2 + hh*kk)*(cc**2) + 3.*(ll**2)*(aa**2))
    )

def diam_spacing(hh, kk, ll, aa):
    return aa / np.sqrt(hh**2 + kk**2 + ll**2)


# ###############################################################
# ###############################################################
#                   Run complement routines
# ###############################################################
# ###############################################################


_complement_dict_mat(_DCRYST_MAT)
_complement_dict_cut(dcryst_mat=_DCRYST_MAT, dcryst_cut=_DCRYST)


def _build_cry(
    crystal=None,
    din=None,
):
    '''
    _build_cry is a function meant to populate a crystal dictionary
    using either hardcoded WEST default crystal names ['Quartz_110', 'Quartz_102']
    or a user-defined crystal as a dictionary in the form of --

    # C-Mod's crystal
    dcry = {
        'material': 'Quartz',
        'name': 'HIREX',
        'symbol': 'Qz102',
        'miller': np.r_[1., 0., 2.,],
        'target': {
            'ion': 'Ar16+',
            'lamb': 3.96,
            'units': 'AA',
            },
        'd_hkl': None,
        }

    '''

    # Hardcoded defaults
    if crystal in ['Quartz_110', 'Quartz_102']:
        dcry = _DCRYST

    # Else if user-defined crystal
    elif isinstance(din, dict):
        dcry = {
            crystal: din
        }

    # Builds the crystal dictionary
    _complement_dict_cut(dcryst_mat=_DCRYST_MAT, dcryst_cut=dcry)

    return dcry[crystal]





# Standard
import numpy as np


# tofu-specific
import tofu.physics_tools as tfpt


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass


def teardown_module():
    pass


#######################################################
#
#     checking Distributions
#
#######################################################


class Test00_Transmission():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_transmission(self):

        # input
        E = np.linspace(0.1, 20, 100) * 1e3
        dthick = {
            'filt1': {
                'mat': 'StainlessSteel',
                'thick': np.r_[5e-6, 10e-6],
            },
            'filt2': {
                'mat': 'Al',
                'thick': np.r_[10e-6, 100e-6],
            },
        }

        # format
        for k0, v0 in dthick.items():
            dthick[k0]['thick'] = v0['thick'][None, :]

        # single
        dout = tfpt.transmission.get_xray_transmission(
            E=E[:, None],
            dthick=dthick,
        )
        assert isinstance(dout, dict)




# -*- coding: utf-8 -*-

# Built-in
import sys
import os
# import itertools as itt
import copy
import warnings
if sys.version[0] == '3':
    import inspect
else:
    # Python 2 back-porting
    import funcsigs as inspect

# Common
import numpy as np

# tofu
from tofu import __version__ as __version__
import tofu.pathfile as tfpf
import tofu.utils as utils
from ._DataCollection_class import DataCollection


__all__ = ['DataCollectPlot']
_SAVEPATH = os.path.abspath('./')


#############################################
#############################################
#       Matplotlib
#############################################
#############################################


class DataCollectionPlot_mpl(DataCollection):

    _DTYPES = {'vline': 'set_xdata',
               'hline': 'set_ydata',
               'xline': 'set_xdata',
               'xline': 'set_xdata',
               'yline': 'set_ydata',
               'line': 'set_data',
               'imshow': 'set_data'}

    _dax = {}
    _dobj = {}

    # ---------------------
    # Complement plot-specific dict
    # ---------------------

    def set_dnmax(self, dnmax={}):
        c0 = isinstance(dnmax, dict)
        c0 = c0 and all([kk in self._dgroup['lkey'] and isinstance(vv, int)
                         for kk, vv in dnmax.items()])
        c0 = c0 and all([kk in dnmax.keys() for kk in self._dgroup['lkey']])
        if not c0:
            msg = "dnmax must be a dict associating an int to each group\n"
            msg += "\t- self.dgroup.keys(): {}\n".format(self._dgroup.keys())
            msg += "\t- dnmax: {}".format(dnmax)
            raise Exception(msg)

        for kk, vv in dnmax.items():
            self._dgroup['dict'][kk]['nmax'] = vv
            for kr in self._dgroup['dict'][kk]['lref']:
                self._dref['dict'][kr]['ind'] = np.zeros((vv,), dtype=int)
                self._dref['dict'][kr]['value'] = np.zeros((vv,), dtype=int)

    # ---------------------
    # Get back to collection
    # ---------------------

    def extract_collection(self, cls=None):
        return cls(dref=self.dref, ddata=self.ddata)

    # ---------------------
    # Set dax, dobj
    # ---------------------

    def set_daxobj(self, dax=None, dobj=None):

        # ----------------------
        # Check basic formatting on dax

        c0 = isinstance(dax, dict)
        c0 = c0 and all([isinstance(kk, str) and isinstance(vv, dict)
                         for kk, vv in dax.items()])
        ls = ['ax', 'lobj', 'xref', 'yref']
        c0 = c0 and all([all([ss in vv.keys() for ss in ls]) for vv in
                         dax.values()])
        if not c0:
            msg = (
                "dax must be a dict with unique keys\n"
                + "Values must be dict with keys:\n"
                + "\t- {}".format(ls)
            )
            raise Exception(msg)

        # ----------------------
        # Check basic formatting on dobj

        c0 = isinstance(dobj, dict)
        c0 = c0 and all([isinstance(kk, str) and isinstance(vv, dict)
                         for kk, vv in dobj.items()])
        ls = ['axe', 'Type', 'refs', 'inds']
        c0 = c0 and all([all([ss in vv.keys() for ss in ls]) for vv in
                         dobj.values()])
        if not c0:
            msg = (
                "dax must be a dict with unique keys\n"
                + "Values must be dict with keys:\n"
                + "\t- {}".format(ls)
            )
            raise Exception(msg)

        # ----------------------
        # Check further formatting on dax

        dax['dict'] = {}
        dax['lkey'] = []
        for k0, v0 in dax.items():
            assert all([k1 in self._ddata['lkey'] for k1 in v0['xref']])
            assert all([k1 in self._ddata['lkey'] for k1 in v0['yref']])
            assert all([k1 in doj.keys() for k1 in v0['lobj']])
            # mpl-specific
            assert isinstance(vv['ax'], )
            dax['dict'][k0] = v0
            del dax[k0]
            dax['lkey'].append(k0)

        # ----------------------
        # Check further formatting on dobj

        dobj['dict'] = {}
        dobj['lkey'] = []
        for k0, v0 in dobj.items():
            assert all([k1 in dax['lkey'] for k1 in v0['axe']])
            assert all([k1 in self._DTYPES.keys() for k1 in v0['Type']])
            assert all([k1 in self._ddata['lkey'] for k1 in v0['data']])
            assert all([k1 in doj.keys() for k1 in v0['inds']])
            # mpl-specific
            assert isinstance(vv['ax'], )
            dobj['dict'][k0] = v0
            del dobj[k0]
            dobj['lkey'].append(k0)

        self._dax.update(dax)
        self._dobj.update(dobj)

    # ---------------------
    # Methods for extracting axes
    # ---------------------

    def get_lax(self, inkey):
        return [kk for kk in self._dax['lkey'] if inkey in kk]

    # ---------------------
    # Methods for showing data
    # ---------------------

    def get_summary(self, data=False, show=None, show_core=None,
                    sep='  ', line='-', just='l',
                    table_sep=None, verb=True, return_=False):
        """ Summary description of the object content """
        # # Make sure the data is accessible
        # msg = "The data is not accessible because self.strip(2) was used !"
        # assert self._dstrip['strip']<2, msg

        # -----------------------
        # Build for groups
        col0 = ['group name', 'nb. ref', 'nb. data', 'nmax']
        ar0 = [(k0,
                len(self._dgroup['dict'][k0]['lref']),
                len(self._dgroup['dict'][k0]['ldata']),
                self._dgroup['dict'][k0]['nmax'])
               for k0 in self._dgroup['lkey']]

        # -----------------------
        # Build for refs
        col1 = ['ref key', 'group', 'size', 'nb. data', 'ind', 'value']
        ar1 = [(k0,
                self._dref['dict'][k0]['group'],
                self._dref['dict'][k0]['size'],
                len(self._dref['dict'][k0]['ldata']),
                str(self._dref['dict'][k0]['ind']),
                str(self._dref['dict'][k0]['value']))
               for k0 in self._dref['lkey']]

        # -----------------------
        # Build for ddata
        if data:
            col2 = ['data key']
            if show_core is None:
                show_core = self._show_in_summary_core
            if isinstance(show_core, str):
                show_core = [show_core]
            lkcore = ['shape', 'groups', 'refs']
            assert all([
                ss in self._ddata['lparam'] + lkcore for ss in show_core
            ])
            col2 += show_core

            if show is None:
                show = self._show_in_summary
            if show == 'all':
                col2 += self._ddata['lparam']
            else:
                if isinstance(show, str):
                    show = [show]
                assert all([ss in self._ddata['lparam'] for ss in show])
                col2 += show

            ar2 = []
            for k0 in self._ddata['lkey']:
                v0 = self._ddata['dict'][k0]
                lu = [k0] + [str(v0[cc]) for cc in col2[1:]]
                ar2.append(lu)

        # -----------------------
        # Build for dax
        if len(self._dax) > 0:
            col3 = ['axes key', 'pos',
                    'xlabel', 'ylabel', 'xlim', 'ylim',
                    'nobj', 'xref / nxref', 'yref / nyref']
            ar3 = [(k0,
                    self._dax['dict'][k0]['ax'].get_position(),
                    self._dax['dict'][k0]['ax'].get_xlabel(),
                    self._dax['dict'][k0]['ax'].get_ylabel(),
                    self._dax['dict'][k0]['ax'].get_xlim(),
                    self._dax['dict'][k0]['ax'].get_ylim(),
                    len(self._dax['dict'][k0]['lobj']),
                    (self._dax['dict'][k0]['xref'][0]
                     + ' / ' + len(self._dax['dict'][k0]['xref']))
                    (self._dax['dict'][k0]['yref'][0]
                     + ' / ' + len(self._dax['dict'][k0]['yref'])))
                   for k0 in self._dax['lkey']]
        else:
            col3 = None
            ar3 = None

        # -----------------------
        # Build for dobj
        if len(self._dobj) > 0:
            col4 = ['obj key', 'axe',
                    'Type', 'refs', 'inds']
            ar4 = [(k0,
                    self._dobj['dict'][k0]['ax'],
                    self._dobj['dict'][k0]['Type'],
                    self._dobj['dict'][k0]['refs'],
                    self._dobj['dict'][k0]['inds'])]
        else:
            col4 = None
            ar4 = None

        # -----------------------
        # concatenate
        lar = [ar0, ar1]
        lcol = [col0, col1]
        if data is True:
            lar.append(ar2)
            lcol.append(col2)
        if ar3 is not None:
            lar.append(ar3)
            lcol.append(col3)
        if ar4 is not None:
            lar.append(ar4)
            lcol.append(col4)

        return self._get_summary(
            lar, lcol,
            sep=sep, line=line, table_sep=table_sep,
            verb=verb, return_=return_,
        )

    # ---------------------
    # Method for getting corresponding nearest index in != ref
    # ---------------------

    def _get_current_ind_for_data(self, key, obj=None):
        if obj is None:
            ind = [self._dref['dict'][kr]['ind']
                   for kr in self._ddata['dict'][key]['refs']]
        else:
            ind = [
                self._dref['dict'][kr]['ind'][
                    self._dobj['dict'][obj]['indi'][kr]
                ]
                for kr in self._ddata['dict'][key]['refs']
            ]
        return ind

    def _get_current_data_along_axis(self, key, axis=None, obj=None):
        dim = self._ddata['dict'][key]['dim']
        if dim == 1:
            return self._ddata['dict'][key]['data']
        else:
            ind = self._get_current_ind_for_data(key)
            if dim == 2:
                if axis == 0:
                    return self._ddata['dict'][key]['data'][:, ind[1]]
                else:
                    return self._ddata['dict'][key]['data'][ind[0]]
            elif dim == 3:
                if axis == 0:
                    return self._ddata['dict'][key]['data'][:, ind[1], ind[2]]
                elif axis == 1:
                    return self._ddata['dict'][key]['data'][ind[0]][:, ind[2]]
                else:
                    return self._ddata['dict'][key]['data'][ind[0]][ind[1]]

    def _get_data_ind_from_value(self, value=None, key=None,
                                 ref=None, axis=None):
        if axis is None:
            axis = self._ddata['dict'][key]['refs'].index(ref)
            size = self._dref['dict'][key]['size']
        else:
            pass

        if self._ddata['dict'][key]['bins'] is not None:
            # bins only exist for 1d sorted data
            ind = np.searchsorted([value], self._ddata['dict'][key]['bins'])[0]
        else:
            x = self._get_current_data_along_axis(key, axis)
            ind = np.nanargmin(np.abs(x-value))
        return ind % size

    def _get_current_value(self, key):
        ind = self._get_current_ind_for_data(key)
        dim = self._ddata['dict'][key]['dim']
        if dim == 1:
            return self._ddata['dict'][key]['data'][ind[0]]
        elif dim == 2:
            return self._ddata['dict'][key]['data'][ind[0]][ind[1]]
        elif dim == 3:
            return self._ddata['dict'][key]['data'][ind[0]][ind[1]][ind[2]]



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 10 21:01:58 2023

@author: dvezinet
"""


# common
import numpy as np
import scipy.constants as scpct
import astropy.units as asunits


__all__ = ['convert_spectral']


_DREF = {
    'length': 'm',
    'frequency': 'Hz',
    'energy': 'J',
}


# #############################################################################
# #############################################################################
#                       Units conversion
# #############################################################################


def convert_spectral(
    data_in=None,
    units_in=None,
    units_out=None,
):
    """ convert wavelength / energy/ frequency
    """

    # -------------
    # Check inputs

    data_in, uin, uout = _check_convert_spectral(
        data_in=data_in,
        units_in=units_in,
        units_out=units_out,
    )

    # -----------
    # compute

    # trivial case first
    if units_in == units_out:
        return data_in

    coef, inv, cat = _convert_spectral_coef(
        uin=uin,
        uout=uout,
    )

    # -----------
    # return

    if data_in is None:
        data = None
    else:
        if inv:
            data = coef / data_in
        else:
            data = coef * data_in

    return data, coef, inv, cat


# ###############################
#       checks
# ###############################


def _check_convert_spectral(
    data_in=None,
    units_in=None,
    units_out=None,
):

    # data_in
    if data_in is None:
        pass
    else:
        if not isinstance(data_in, np.ndarray):
            try:
                data_in = np.asarray(data_in)
            except Exception as err:
                msg = (
                    "Arg data shall be convertible to a np.ndarray!\n"
                    + str(err)
                )
                raise Exception(msg)
        if data_in.dtype not in [int, float]:
            msg = (
                "Arg data must be a np.ndarray of dtype int or float!\n"
                f"data.dtype = {data_in.dtype.name}"
            )
            raise Exception(msg)

    # units
    uin = asunits.Unit(units_in)
    uout = asunits.Unit(units_out)

    ltypes = ['energy', 'length', 'frequency']

    c0 = any([ss in uin.physical_type for ss in ltypes])
    if not c0:
        msg = (
            "units_in is not recognized as a relevant spectral quantity:\n"
            f"\t- units_in: {units_in}\n"
            f"\t- physical_type: {uin.physical_type}\n"
            f"\t- relevant physical types: {ltypes}"
        )
        raise Exception(msg)

    c0 = any([ss in uout.physical_type for ss in ltypes])
    if not c0:
        msg = (
            "units_out is not recognized as a relevant spectral quantity:\n"
            f"\t- units_out: {units_out}\n"
            f"\t- physical_type: {uout.physical_type}\n"
            f"\t- relevant physical types: {ltypes}"
        )
        raise Exception(msg)

    return data_in, uin, uout


# ###############################
#       compute
# ###############################


def _convert_spectral_coef(
    uin=None,
    uout=None,
):
    """ Get conversion coef """

    ltypes = ['energy', 'length', 'frequency']
    k0_in = [ss for ss in ltypes if ss in uin.physical_type][0]
    k0_out = [ss for ss in ltypes if ss in uout.physical_type][0]

    if uin == uout:
        return 1., False

    # ---------
    # First case: same physical type

    inv = False
    if k0_in == k0_out:
        coef = uin.in_units(uout)

    # ---------
    # For each category, convert to reference (m, eV, Hz)
    else:

        # coefs_in
        coef_in = uin.in_units(_DREF[k0_in])

        # coefs_out
        coef_out = uin.in_units(_DREF[k0_out])

        # ------------------
        # Cross combinations between (m, J, Hz)

        # E = h*f = h*c/lambda
        if k0_in == 'length':
            inv = True
            if k0_out == 'energy':
                # m -> J
                coef_cross = scpct.h * scpct.c
            elif k0_out == 'frequency':
                # m -> Hz
                coef_cross = scpct.c

        elif k0_in == 'energy':
            if k0_out == 'length':
                # J -> m
                inv = True
                coef_cross = scpct.h * scpct.c
            elif k0_out == 'frequency':
                # J -> Hz
                coef_cross = 1./scpct.h

        elif k0_in == 'frequency':
            if k0_out == 'length':
                # Hz -> m
                inv = True
                coef_cross = scpct.c
            elif k0_out == 'energy':
                # Hz -> J
                coef_cross = scpct.h

        if inv:
            coef = coef_cross*coef_out / coef_in
        else:
            coef = coef_in*coef_cross*coef_out

    return coef, inv, k0_out


"""
Creating and using diagnostic
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

"""


import os
import copy
import numpy as np


import tofu as tf


_PATH_HERE = os.path.dirname(__file__)


__all__ = ['main']


# #####################################################
# #####################################################
#           Main
# #####################################################


def main():

    # ------------------------
    # create plasma

    conf, coll = _create_plasma()

    # ------------------------
    # add several diagnostics

    # add broadband
    _add_broadband(coll, key_diag='d0', conf=conf, vos=True)

    # add collimator
    # _add_collimator(coll, conf, vos=True)

    # add 2d camera
    _add_2d(coll, key_diag='c2d', conf=conf, vos=True)

    # add PHA
    # _add_PHA(coll, conf)

    # add spectrometer
    # _add_spectrometer(coll, conf, vos=True)   # , crystals=['c0'])

    # add spectro-like without crystal
    # _add_spectrometer_like(coll, config=conf, key_diag='d02')

    # ------------------------
    # compute synthetic signal

    _compute_synth_signal(
        coll,
        ldiag=['d0', 'c2d'],
        method='vos',
        spectral_binning=True,
    )

    # ------------------
    # geometry matrices

    # ----------
    # inversions

    return conf, coll


# #####################################################
# #####################################################
#           Routines - PLasma
# #####################################################


def _create_plasma():

    # -----------------
    # load a simple geometry

    conf = tf.load_config('WEST-V0')

    # -----------------
    # Instanciate a collection

    coll = tf.data.Collection()

    # -----------------
    # add a rect mesh with 2d bsplines

    coll.add_mesh_2d_rect(
        key='m0',
        res=0.10,
        crop_poly=conf,
        units='m',
        deg=1,
    )

    # coll.add_mesh_2d_rect(
        # key='m1',
        # res=0.03,
        # crop_poly=conf,
        # units='m',
        # deg=0,
    # )

    # --------
    # add time

    nt = 11
    t = np.linspace(0, 10, nt)

    coll.add_ref('nt', nt)
    coll.add_data('t', data=t, ref='nt', units='s', dime='time')

    # ------------------
    # add a spectra mesh

    nE = 100
    E = np.linspace(1000, 30000, nE)

    coll.add_mesh_1d(
        key='mE',
        knots=E,
        deg=1,
        units='eV',
    )

    # ---------------------
    # get shape of bs coefs to set a time-vaying 2d radius

    kapR, kapZ = coll.dobj['bsplines']['m0_bs1']['apex']
    apR = coll.ddata[kapR]['data']
    apZ = coll.ddata[kapZ]['data']

    # rho2d
    rho2d = (
        (1 + 0.1*np.cos(t[:, None, None]))
        * (1. - np.exp(
            -(apR[None, :, None]-2.5)**2/0.4**2
            -(apZ[None, None, :]-0)**2/0.6**2
        ))
    )

    coll.add_data(
        'rho2d',
        data=rho2d,
        ref=('nt', 'm0_bs1'),
        units='',
    )

    # add 1d radial mesh based on rho2d
    nrho = 20
    rho = np.linspace(0, 1, nrho)
    coll.add_mesh_1d(
        key='mr',
        knots=rho,
        units='',
        subkey='rho2d',
        deg=1,
    )

    # emiss1dE
    r0 = (0. + np.exp(-rho**2/0.4**2)[None, :, None])
    r1 = np.exp(-(rho-0.3)**2/0.15**2)[None, :, None]
    emiss1dE = (
        (1 + 0.1*np.cos(t[:, None, None]))
        * (
            r0 * (0.1 + np.exp(-E/10000))[None, None, :]
            + r1 * np.exp(-(E-15000)**2/1000**2)[None, None, :]
        )
    )

    coll.add_data(
        'emiss1d',
        data=np.sum(emiss1dE, axis=-1),
        ref=('nt', 'mr_bs1'),
        units='ph/(m3.s.sr)',
    )

    # emiss1dE
    coll.add_data(
        'emiss1dE',
        data=emiss1dE,
        ref=('nt', 'mr_bs1', 'mE_bs1'),
        units='ph/(m3.s.sr.eV)',
    )

    # emiss2dE
    r0 = (0. + np.exp(-rho2d**2/0.4**2)[:, :, :, None])
    r1 = np.exp(-(rho2d-0.3)**2/0.15**2)[:, :, :, None]
    emiss2dE = (
        (1 + 0.1*np.cos(t[:, None, None, None]))
        * (
            r0 * (1. + np.exp(-E/10000))[None, None, None, :]
            + r1 * np.exp(-(E-15000)**2/1000**2)[None, None, None, :]
        )
    )

    coll.add_data(
        'emiss2dE',
        data=emiss2dE,
        ref=('nt', 'm0_bs1', 'mE_bs1'),
        units='ph/(m3.s.sr.eV)',
    )

    return conf, coll


# #####################################################
# #####################################################
#           Routines - Diags
# #####################################################


def _add_broadband(
    coll=None,
    key_diag=None,
    conf=None,
    vos=None,
):

    # ---------------------
    # add 2 pinhole cameras

    # coll.add_camera_pinhole(
        # key='bb0',
        # key_pinhole=None,
        # key_diag=key_diag,
        # cam_type='1d',
        # R=3.3,
        # z=-0.6,
        # phi=0,
        # theta=3.*np.pi/4,
        # dphi=np.pi/10,
        # tilt=np.pi/2,
        # focal=0.1,
        # pix_nb=10,
        # pix_size=3e-3,
        # pix_spacing=5e-3,
        # pinhole_radius=None,
        # pinhole_size=[2e-3, 1e-3],
        # reflections_nb=0,
        # reflections_type=None,
        # compute=False,
        # config=conf,
    # )

    coll.add_camera_pinhole(
        key='bb1',
        key_pinhole=None,
        key_diag=key_diag,
        cam_type='1d',
        R=3.3,
        z=0.6,
        phi=0,
        theta=-3.*np.pi/4,
        dphi=np.pi/10,
        tilt=np.pi/2,
        focal=0.1,
        pix_nb=10,
        pix_size=3e-3,
        pix_spacing=5e-3,
        pinhole_radius=None,
        pinhole_size=[3e-3, 2e-3],
        reflections_nb=0,
        reflections_type=None,
        compute=True,
        config=conf,
    )

    if vos is True:
        coll.compute_diagnostic_vos(
            key_diag=key_diag,
            key_mesh='m0',
            res_RZ=0.005,
            res_phi=0.01,
            visibility=False,
            store=True,
        )

    return


def _add_collimator(
    coll=None,
    conf=None,
    vos=None,
):

    # ---------------------
    # add 2 pinhole cameras

    # aperture 0
    dgeom = {
        'cent': np.r_[3.3, 0, 0],
        'nin': np.r_[-1, 0, 0],
        'e0': np.r_[0, -1, 0],
        'e1': np.r_[0, 0, 1],
        'outline_x0': 0.005 * np.r_[-1, 1, 1, -1],
        'outline_x1': 0.005 * np.r_[-1, -1, 1, 1],
    }

    coll.add_aperture(
        key='coll_ap00',
        **dgeom,
    )

    # add camera
    del dgeom['cent']
    dgeom['cents_x'] = 3.5 * np.r_[1, 1, 1]
    dgeom['cents_y'] = np.r_[0, 0, 0]
    dgeom['cents_z'] = np.r_[-0.05, 0, 0.05]

    coll.add_camera_1d(
        key='coll_cam',
        dgeom=dgeom,
    )

    # create diagnostic
    coll.add_diagnostic(
        key='coll',
        doptics={
            'coll_cam': [
                ('coll_ap00',),
                ('coll_ap00',),
                ('coll_ap00',),
            ],
        },
        compute=True,
    )

    if vos is True:
        coll.compute_diagnostic_vos(
            'dcoll',
            key_mesh='m0',
            res_RZ=0.01,
            res_phi=0.01,
            visibility=False,
            store=True,
        )

    return

def _add_2d(
    coll=None,
    key_diag=None,
    conf=None,
    vos=None,
):

    # ---------------------
    # add 2 pinhole cameras

    coll.add_camera_pinhole(
        key=key_diag,
        key_diag=key_diag,
        key_pinhole=None,
        cam_type='2d',
        R=3.3,
        z=-0.6,
        phi=0,
        theta=3.*np.pi/4,
        dphi=np.pi/6,
        tilt=np.pi/2,
        focal=0.1,
        pix_nb=[5, 3],
        pix_size=1e-3,
        pix_spacing=5e-3,
        pinhole_radius=5e-3,
        pinhole_size=None,
        reflections_nb=0,
        reflections_type=None,
        compute=True,
        config=conf,
    )

    if vos is True:
        coll.compute_diagnostic_vos(
            key_diag,
            key_mesh='m0',
            res_RZ=0.02,
            res_phi=0.02,
            visibility=False,
            store=True,
        )


def _add_PHA(
    coll=None,
    conf=None,
):

    nEb = 14
    Ebins = np.linspace(4000, 20000, nEb)

    # Single pinhole cam
    coll.add_camera_pinhole(
        key='pha0',
        key_pinhole=None,
        key_diag='d2',
        cam_type='1d',
        R=5,
        z=0,
        phi=0,
        theta=np.pi,
        dphi=0,
        tilt=0,
        focal=0.1,
        pix_nb=5,
        pix_size=1e-2,
        pix_spacing=0,
        pinhole_radius=1e-3,
        pinhole_size=None,
        reflections_nb=0,
        reflections_type=None,
        compute=True,
        config=conf,
        dmat={'bins': Ebins},
    )

    return


def _add_spectrometer_like(
    coll=None,
    config=None,
    key_diag=None,
):

    # --------
    # aperture

    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    key_cam = list(doptics.keys())[0]
    kcryst, kslit = doptics[key_cam]['optics']

    # ------------
    # flat crystal

    # dgeom
    dgeom = copy.deepcopy({
        k0: coll.dobj['crystal'][kcryst]['dgeom'][k0]
        for k0 in ['cent', 'nin', 'e0', 'e1']
    })
    dgeom['curve_r'] = [np.inf, np.inf]
    dgeom['extenthalf'] = 0.2 * np.r_[1, 1]

    # dmat
    dmat = copy.deepcopy({
        k0: coll.dobj['crystal'][kcryst]['dmat'][k0]
        for k0 in ['d_hkl', 'target', 'drock']
    })

    for k0 in ['angle_rel', 'power_ratio']:
        dmat['drock'][k0] = coll.ddata[dmat['drock'][k0]]['data']

    coll.add_crystal(
        key='c0_flat',
        dgeom=dgeom,
        dmat=dmat,
    )

    # ---------------
    # camera replica

    # dgeom
    dgeom = copy.deepcopy({
        k0: coll.dobj['camera'][key_cam]['dgeom'][k0]
        for k0 in ['cent', 'nin', 'e0', 'e1']
    })
    kout0, kout1 = coll.dobj['camera'][key_cam]['dgeom']['outline']
    dgeom['outline_x0'] = coll.ddata[kout0]['data']
    dgeom['outline_x1'] = coll.ddata[kout1]['data']
    cx0, cx1 = coll.dobj['camera'][key_cam]['dgeom']['cents']
    dgeom['cents_x0'] = coll.ddata[cx0]['data']
    dgeom['cents_x1'] = coll.ddata[cx1]['data']

    # add camera replica
    coll.add_camera_2d(
        key='c0_camf',
        dgeom=dgeom,
    )

    # ---------------
    # camera rotated

    dgeom = copy.deepcopy(dgeom)

    # bragg angle
    cent = coll.dobj['crystal'][kcryst]['dgeom']['cent']
    nin = coll.dobj['crystal'][kcryst]['dgeom']['nin']
    e0 = coll.dobj['crystal'][kcryst]['dgeom']['e0']
    e1 = coll.dobj['crystal'][kcryst]['dgeom']['e1']
    ang = 2*coll.get_crystal_bragglamb('c0')[0]

    # rotate cent

    dv = (dgeom['cent'] - cent)
    scain = np.sum(dv * nin)
    sca0 = np.sum(dv * e0)
    sca1 = np.sum(dv * e1)
    assert np.abs(sca1) < 1e-8

    dvbis = (
        np.cos(ang) * (scain * nin + sca0 * e0)
        + np.sin(ang) * (sca0 * nin - scain * e0)
        + sca1 * e1
    )
    dgeom['cent'] = cent + dvbis

    # unite vectors
    for k0 in ['nin', 'e0', 'e1']:
        scain = np.sum(dgeom[k0] * nin)
        sca0 = np.sum(dgeom[k0] * e0)
        sca1 = np.sum(dgeom[k0] * e1)
        dgeom[k0] = (
            np.cos(ang) * (scain * nin + sca0 * e0)
            + np.sin(ang) * (sca0 * nin - scain * e0)
            + sca1 * e1
        )

    # add rotated camera
    coll.add_camera_2d(
        key='c0_cam1',
        dgeom=dgeom,
    )

    # -----------
    # diagnostic

    # flat crystal
    coll.add_diagnostic(
        doptics={'c0_camf': ['c0_flat', kslit]},
        config=config,
        compute=True,
    )

    # no crystal
    coll.add_diagnostic(
        doptics={'c0_cam1': [kslit]},
        config=config,
        compute=True,
    )


def _add_spectrometer(
    coll=None,
    conf=None,
    crystals=None,
    vos=None,
):

    # ------------------
    # add crystal optics

    doptics = {}
    dcrystals = _crystals(coll, crystals=crystals)

    # ----------------------
    # add camera / apertures

    for k0, v0 in dcrystals.items():

        if k0 != 'c0':
            continue

        loptics = coll.get_crystal_ideal_configuration(
            key=k0,
            configuration=v0['configuration'],
            # parameters
            cam_on_e0=False,
            cam_tangential=True,
            cam_dimensions=np.r_[1028, 512]*75e-6,
            focal_distance=2.,
            defocus=0.,
            # defocus=-1.5,
            # store
            store=True,
            key_cam=f'{k0}_cam',
            aperture_dimensions=[100e-6, 1e-2],
            pinhole_radius=100e-6 if v0['configuration'] == 'pinhole' else None,
            cam_pixels_nb=[21, 11],
            # cam_pixels_nb=[41, 41],
            # returnas
            returnas=list,
        )

        # add diag
        gtype = coll.dobj['crystal'][k0]['dgeom']['type']
        if gtype not in ['toroidal']:
            doptics.update({
                k0: loptics,
            })

    # ------------------
    # add diagnostic

    for k0, v0 in doptics.items():
        coll.add_diagnostic(
            doptics=v0,
            config=conf,
            compute=True,
            add_points=3,
            rocking_curve_fwhm=0.0001*np.pi/180 if k0 == 'c2' else None,
        )

        if vos is True:
            coll.compute_diagnostic_vos(
                list(coll.dobj['diagnostic'].keys())[-1],
                key_mesh='m0',
                res_RZ=[0.10, 0.01],
                res_phi=0.005,
                res_lamb=0.001e-10,
                n0=11,
                n1=21,
                store=True,
            )

    return


def _crystals(coll=None, crystals=None):

    # -------
    # check

    if crystals is None:
        crystals = ['c0', 'c1', 'c2']

    # -------
    # geom

    start, vect, v0, v1 = _ref_line(
        start=np.r_[17.918, -2.157, 0.043],
        vect=np.r_[-0.29770273, 0.95465862, 0.],
    )
    # start, vect, v0, v1 = _ref_line(start=np.r_[7., 0, 0.001])

    # cryst0: planar
    cent = start + 0. * vect

    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        # theta=-np.pi/4,
        theta=0,
        phi=0.,
    )

    dc = {}

    # c1: cylindrical (von hamos)
    if 'c0' in crystals:

        # load rocking curve
        pfe = os.path.join(_PATH_HERE, 'Ge242.txt')
        out = np.loadtxt(pfe)
        drock = {
            'angle_rel': out[:, 0],
            'power_ratio': out[:, 1],
        }

        size = 1.e-2
        rc = 1.03

        c0 = {
            'key': 'c0',
            'dgeom': {
                'cent': cent,
                'nin': nin,
                'e0': e0,
                'e1': e1,
                'extenthalf': size * np.r_[1, 1/rc],
                'curve_r': [np.inf, rc],
            },
            # 'dmat': 'Quartz_110',
            'dmat': {
                'material': 'Germanium',
                'name': 'Ge224',
                'miller': np.r_[2,2,4],
                # 'd_hkl': 0.944e-10 / (2*np.sin(24.2*np.pi/180.)),
                'target': {'lamb': 0.944e-10},
            },
            'configuration': 'von hamos',
        }
        dc['c0'] = c0
        coll.add_crystal(c0['key'], dgeom=c0['dgeom'], dmat=c0['dmat'])

    # c3: cylindrical (convex)
    if 'c1' in crystals:
        rc = 2.
        ang = np.linspace(-0.0001, 0.0005, 100)
        c1 = {
            'key': 'c1',
            'dgeom': {
                'cent': cent,
                'nin': nin,
                'e0': e0,
                'e1': e1,
                'extenthalf': size * np.r_[1/rc, 1],
                'curve_r': [-rc, np.inf],
            },
            'dmat': {
                'target': {'lamb': 3.94e-10},
                'd_hkl': 2.45652e-10,
                'drock': {
                    'angle_rel': ang,
                    'power_ratio': np.exp(-(ang-0.00025)**2/0.0001**2),
                },
            },
            'configuration': 'pinhole',
        }
        dc['c1'] = c1
        coll.add_crystal(c1['key'], dgeom=c1['dgeom'], dmat=c1['dmat'])

    # c2: spherical
    if 'c2' in crystals:
        rc = 2.
        c2 = {
            'key': 'c2',
            'dgeom': {
                'cent': cent,
                'nin': nin,
                'e0': e0,
                'e1': e1,
                'extenthalf': size * np.r_[1 / rc, 1 / rc],
                'curve_r': rc,
            },
            'dmat': 'Quartz_110',
            'configuration': 'johann',
        }
        dc['c2'] = c2
        coll.add_crystal(c2['key'], dgeom=c2['dgeom'], dmat=c2['dmat'])

    return dc


def _ref_line(start=np.r_[4, 0, 0], vect=np.r_[-1, 0, 0]):

    vect = vect / np.linalg.norm(vect)

    v0 = np.r_[-vect[1], vect[0], 0.]
    v0 = v0 / np.linalg.norm(v0)

    v1 = np.cross(vect, v0)

    return start, vect, v0, v1


def _nine0e1_from_orientations(
    vect=None,
    v0=None,
    v1=None,
    theta=None,
    phi=None,
):

    nin = (
        vect * np.cos(theta)
        + np.sin(theta) * (np.cos(phi) * v0 + np.sin(phi) * v1)
    )

    e0 = (
        - vect * np.sin(theta)
        + np.cos(theta) * (np.cos(phi) * v0 + np.sin(phi) * v1)
    )

    e1 = np.cross(nin, e0)

    return nin, e0, e1


# #####################################################
# #####################################################
#           Synthetic signal
# #####################################################


def _compute_synth_signal(
    coll=None,
    ldiag=None,
    method=None,
    spectral_binning=None,
):

    # -------------
    # list of diags

    if ldiag is None:
        ldiag = list(coll.dobj['diagnostic'])

    # ------------
    # loop n diags

    for k0 in ldiag:

        if k0 not in coll.dobj['diagnostic'].keys():
            continue

        # params
        if k0 in ['d0', 'c2d']:
            key_integrand = 'emiss1d'
            ref_com = 'nt'
        # elif k0 == 'diag00':
            # key_integrand = 'emiss2dE'
            # ref_com = None
        else:
            key_integrand = 'emiss1dE'
            ref_com = 'nt'

        # compute
        coll.compute_diagnostic_signal(
            key=None,
            key_diag=k0,
            key_cam=None,
            key_integrand=key_integrand,
            method=method,
            res=0.001,
            mode='abs',
            groupby=None,
            val_init=None,
            ref_com=ref_com,
            brightness=None,
            spectral_binning=spectral_binning,
            verb=True,
            timing=False,
            store=True,
            returnas=False,
        )


# #####################################################
# #####################################################
#           Script
# #####################################################


if __name__ == '__main__':
    main()
